/* -----------------------------------------
 Syntax translator from INDY to Curry
 Santiago Escobar <sescobar@dsic.upv.es>
 November 1998 Valencia
 ELP research group
 ----------------------------------------- */

:- use_module(library(system)).

/*-----------------------------
 Type checking for translated program in Curry is done 
 by introducing a general purpose type
 that has all constructors of program, as logic's Herbrand's base.
 -----------------------------*/
:- dynamic constructor/2. % name and arity

% --------------------------------------
% Convert from Indy syntax to typed Curry syntax
indy2curry( FileIndy ) :-
	retractall( constructor(_,_) ),
	readFile( FileIndy,StrIndy ),
	translate( StrIndy, [], StrCurry ),
	gen_type( TypeCurry ),
	append( TypeCurry, StrCurry, Curry ),
	name( FileIndy, FileIndyStr ),
        append( FileIndyStr, ".curry", FileCurryStr ),
	name( FileCurry, FileCurryStr ),
	writeFile( FileCurry, Curry ).

% --------------------------------------
% Translate directly from Indy to Curry
% 1 arg -> input string
% 2 arg -> output string
% 3 arg -> output Curry string
% (same in general for all translation predicates)
translate( [], [], [] ).
translate( In, Out, OutC ) :-	
	trans_rule( In, In2, Out2C ),
	append( Out2C, Out3C, OutC ),
	translate( In2, Out, Out3C ).

% Translate rule of indy program to syntax Curry (OutC)
% Unique rule shape:  term --> term,...,term.
trans_rule( In, Out, OutC ) :-
	trans_term( nopar, In, In2, TermC ),
	In2 = [45,62|In3],		% "->" == 45,62
	trans_terms( nopar, 32, In3, In4, TermsC, _ ),
	In4 = [59|Out],				% ";"  == 59
	append( TermC, " = ", Aux1 ),
	append( Aux1, TermsC, Aux2 ),
	append( Aux2, "\n", OutC ).

% Extract a curry term from input term
trans_term( Par, In, Out, OutC ) :-
	In = [Ch|In2],
	(
	  97 =< Ch, Ch =< 122,!,	% ---function
	  trans_name( Str, In, In3 ), 
	  ( % n-arity function 
	    In3 = [40|In4],!,			% "(" == 40, ")" == 41
	    trans_terms( par, 32, In4, In5, TermsC, _ ),
	    In5 = [41|Out],
	    append( Str, " ", Aux1 ),
	    append( Aux1, TermsC, Aux2 ),
	    ( Par = nopar, OutC = Aux2
	     ;Par = par, append( "(", Aux2, Aux3), append( Aux3, ")", OutC)  )
	   ;% 0-arity function
	    Out = In3,
	    OutC = Str   )

	 ;Ch = 58,!,			% ---constructor
	  trans_name( Str1, In2, In3 ),
	  % checks and correct the first character of string
	  check_nonumber( Str1, Str ),
	  ( % n-arity constructor
	    In3 = [40|In4],!,  
	    trans_terms( par, 32, In4, In5, TermsC, Narg ),
	    In5 = [41|Out],
	    append( Str, " ", Aux1 ),
	    append( Aux1, TermsC, Aux2 ),
	    ( Par = nopar, OutC = Aux2
	     ;Par = par, append( "(", Aux2, Aux3), append( Aux3, ")", OutC)  )
	   ;% 0-arity constructor
	    Out = In3,
	    OutC = Str,
	    Narg = 0
	  ),
	  ( constructor(Str,Narg)
	   ;\+(constructor(Str,Narg)),
	    assert(constructor(Str,Narg)) )
	    
	 ;65 =< Ch, Ch =< 90,!,		% ---var
	  trans_name( Str, In, Out ),
	  OutC = Str
	  
	 ;In=[91,93|Out],!,			% ---empty list
	  OutC = "[]"

	 ;91 = Ch,!,			% ---list
	  trans_list( Par, In, Out, OutC )

	).

% changes the first character if it's not correct
check_nonumber( [X|Xs], [X|Xs] ) :-
	( 97 =< X, X =< 122
	 ;65 =< X, X =< 90  ), !.
check_nonumber( [X|Xs], [122,X|Xs] ).

% Process a sequence of terms ended either by ")" or ";"
% There's a special arg called Par which controls
% to put or not parenthesis at the Curry term expression.
% The arg Char contains the character code to use as 
% term separator.
trans_terms( Par, Char, In, Out, TermsC, NArgs ) :-
	trans_term( Par, In, In2, TermC ),

	( In2 = [44|In3],!,	% more terms	% "," = 44
	  append( TermC, [Char], Aux1 ),
	  append( Aux1, Terms2C, TermsC ),
	  trans_terms( Par, Char, In3, Out, Terms2C, NArgs2 ),
	  NArgs is NArgs2 + 1

	 ;TermsC = TermC, Out = In2, NArgs = 1 ).

% extract list of valid characters to get a name
trans_name( [X|Name], [X|In], Out ) :-
	( 97 =< X, X =< 122
	 ;65 =< X, X =< 90
	 ;48 =< X, X =< 57
	 ;95 = X ),!,
	 trans_name( Name, In, Out ).
trans_name( [], In, In ).

% translate to Curry according to functional lists
% [1,2,3] -> [1,2,3]
% [1,2|X] -> 1:2:X
trans_list( Par, [91|In], Out, OutC ) :-                % "[" == 91
	( % complete list
	  trans_terms( par, 44, In, In2, Out1C, _ ),	% ":" == 58
	  In2 = [93|Out],  %-- complete list		% "]" == 93
	  append( "[", Out1C, Aux1 ),
	  append( Aux1, "]", OutC )
	 ;%list with tail
	  trans_terms( par, 58, In, In2, Out1C, _ ),	% ":" == 58
	  In2 = [124|In3], %-- list with tail		% "|" == 124
	  trans_term( par, In3, In4, Out2C ),
	  In4 = [93|Out],
	  ( Par = nopar, 
	    append( Out1C, ":", Aux1 ),
	    append( Aux1, Out2C, OutC )
	   ;Par = par,
	    append( "(", Out1C, Aux1 ),
	    append( Aux1, ":", Aux2 ),
	    append( Aux2, Out2C, Aux3 ),
	    append( Aux3, ")", OutC ) )
	).

% ------------------------------
% Clean unnecessary characters
clean( [10|Xs], Ys ) :- !,clean(Xs,Ys).  %intro
clean( [13|Xs], Ys ) :- !,clean(Xs,Ys).  %intro
clean( [32|Xs], Ys ) :- !,clean(Xs,Ys).  %blank
clean( [9|Xs], Ys ) :- !,clean(Xs,Ys).   %tab
clean( [X|Xs], [X|Ys] ) :- clean(Xs,Ys).
clean( [], [] ).

% -------------------------
% clean all rewriting rules
clean_rewriting( [], [] ) :- !.
clean_rewriting( In, Out ) :-
	trans_term( par, In, In2, _ ),
	( In2 = [45,45,62|In3], 	% rewriting rule
	  skipToend( In3, In4 ),
	  clean_rewriting( In4, Out )
	 ;In2 = [45,62|In3],	% normal rule
	  cutToend( Rule, In, In4 ),
	  append( Rule, In5, Out ),
	  clean_rewriting( In4, In5 )
	).

% skip all characters until end of rule
skipToend( [59|Xs], Xs ) :- !.
skipToend( [_|Xs], Ys ) :- skipToend( Xs, Ys ).

% cut rule from whole text (until ;)
cutToend( [59],   [59|Xs], Xs ) :- !.
cutToend( [X|Cs], [X|Xs],  Ys ) :- cutToend( Cs, Xs, Ys ).

% ------------------------------
% Generate general purpose data type
gen_type( "" ) :- \+(constructor(_,_)),!.

gen_type( TypeStr ) :-
	findall(constructor(X,Y),constructor(X,Y),Cons),
	append( "data t = ", ConsStr, Aux1 ),
	append( Aux1, "\n\n", TypeStr ),
	gen_cons( Cons, [], ConsStr ).

% generate sequence of constructors
gen_cons( [constructor(Name,N)], ConsStr1, ConsStr2 ) :-
	append( ConsStr1, " ", Aux1 ),
	append( Aux1, Name, Aux2 ),
	append( Aux2, " ", Aux3 ),
	gen_some( N, Types ),
	append( Aux3, Types, ConsStr2 ).
gen_cons( [constructor(Name,N),X|Cons], ConsStr1, ConsStr3 ) :-
	append( ConsStr1, " ", Aux1 ),
	append( Aux1, Name, Aux2 ),
	append( Aux2, " ", Aux3 ),
	gen_some( N, Types ),
	append( Aux3, Types, Aux4 ),
	append( Aux4, "|", ConsStr2 ),
	gen_cons( [X|Cons], ConsStr2, ConsStr3 ).

% Generate some "t " arguments
gen_some( 0, [] ).
gen_some( N, Str ) :- N > 0,
	append( "t ", Rest, Str ),
	N1 is N - 1,
	gen_some( N1, Rest ).

% ------------------------------
% Read whole file in a variable
readFile( File, Str ) :-
        (  file_exists(File),!
         ; format('ERROR: file ~a doesn\'t exist\n',[File]), fail ),
        (  open( File, read, Stream ),
           read_all( StrDirty, Stream ),
           close( Stream ),!
         ; format('ERROR: file ~a cannot be read\n',[File]), fail ),
        (  clean( StrDirty, StrR ),
           clean_rewriting( StrR, Str),!
         ; format('ERROR: file ~a cannot be clean\n',[File]), fail ).

read_all( Text, Stream ) :-
        get0( Stream, X ),
        ( X =\= -1, Text = [X|Text2], read_all( Text2, Stream )
         ;X =:= -1, Text = [] ).

% ------------------------------
% Write text variable to file
writeFile( File, Str ) :-
         (  open( File, write, Stream ),
	      name( Text, Str ),
            write( Stream, Text ),
            close( Stream ),!
          ; format('ERROR: file ~a cannot be written\n',[File]), fail ).

% -- append
append( [],     Ys, Ys ).
append( [X|Xs], Ys, [X|Zs] ) :- append( Xs, Ys, Zs ).
